;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "outline-minor-faces" "1.2.4"
  "Highlight only section headings."
  '((emacs  "28.1")
    (compat "31.0"))
  :url "https://github.com/tarsius/outline-minor-faces"
  :commit "71705d0708459f765403968567a329a76c38cf62"
  :revdesc "v1.2.4-0-g71705d070845"
  :keywords '("faces" "outlines")
  :authors '(("Jonas Bernoulli" . "emacs.outline-minor-faces@jonas.bernoulli.dev"))
  :maintainers '(("Jonas Bernoulli" . "emacs.outline-minor-faces@jonas.bernoulli.dev")))
